/* Powic: Power Index Computation */
/* 2011/3/31 Takeaki Uno   e-mail:uno@nii.jp, 
    homepage:   http://research.nii.ac.jp/~uno/index.html  */
/* This program is available for only academic use, basically.
   Anyone can modify this program, but he/she has to write down 
    the change of the modification on the top of the source code.
   Neither contact nor appointment to Takeaki Uno is needed.
   If one wants to re-distribute this code, do not forget to 
    refer the newest code, and show the link to homepage of 
    Takeaki Uno, to notify the news about POWIC for the users.
   For the commercial use, please make a contact to Takeaki Uno. */



#ifndef _powic_c_
#define _powic_c_

#include<gmp.h>
#include"stdlib2.c"
#include"aheap.c"

#define POWIC_BZ 4   // Banzhafindex
#define POWIC_SS 8   // Shapley-Shubik index
#define POWIC_DP 16  // Deegan-Packel index


void POWIC_error (){
  ERROR_MES = "POWIC message print";
  print_err ("## POWIC: fast computation of power indices for all players\n\
powic BSD input-filename quota\n\
B:Banzhaf index, S:Shaplay Shubik index, D:Deegan Packel index\n\
if quota is 0, it is set to be the half of all weights, (half+0.5 if odd)\n");
  EXIT;
}

char POWIC_sep = ' ', *POWIC_infname, *POWIC_outfname;
int POWIC_players = 0, POWIC_quota = 0, POWIC_problem=0, POWIC_prec=16;
LONG POWIC_cnt=0, POWIC_cnt_all=0;
mpz_t *POWIC_comb, POWIC_fact;
mpf_t POWIC_minc;
AHEAP POWIC_heap;
double *POWICD_comb =NULL;

/***********************************************************************/
/*  read parameters given by command line  */
/***********************************************************************/
void POWIC_read_param (int argc, char *argv[]){
  int c=1;
  if ( argc < c+3 ){ POWIC_error (); return; }
  
  if ( !strchr (argv[c], 'q') ) POWIC_problem |= SHOW_MESSAGE;
  if ( strchr (argv[c], 'B') ) POWIC_problem |= POWIC_BZ;
  if ( strchr (argv[c], 'S') ) POWIC_problem |= POWIC_SS;
  if ( strchr (argv[c], 'D') ) POWIC_problem |= POWIC_DP;
  c++;
  
  while ( argv[c][0] == '-' ){
    switch (argv[c][1]){
      case ',': POWIC_sep = argv[c+1][0];
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+1 ){ POWIC_error (); return; }
  }

  NEXT:;
  POWIC_infname = argv[c];
  POWIC_quota = atoi(argv[c+1]);
  if ( argc>c+2 ) POWIC_outfname = argv[c+2];
}


/* initialize a line (allocate memory) */
mpz_t *POWIC_mk_line (){
  mpz_t *f;
  int i;
  malloc2 (f, POWIC_quota+2, EXIT);
  FLOOP (i, 0, POWIC_quota+2) mpz_init_set_ui (f[i], 0);
  return (f);
}
mpz_t **POWIC_mk_lines (){
  mpz_t **f;
  int i;
  malloc2 (f, POWIC_players, EXIT);
  FLOOP (i, 0, POWIC_players) f[i] = POWIC_mk_line ();
  return (f);
}

/* clear (free) the allocated memory for one line */
void POWIC_clear_line (mpz_t *f){
  int i;
  FLOOP (i, 0, POWIC_quota+2) mpz_clear (f[i]);
  free2 (f);
}
void POWIC_clear_lines (mpz_t **f){
  int i;
  FLOOP (i, 0, POWIC_players) POWIC_clear_line (f[i]);
  free2 (f);
}

/* compute the next line, and set f to the next line, if ff=NULL.
   if ff is given, regard f=f(*,k), and ff=(*,k+1) (considering #players version) */
/* dir=1 ==> compute f[i+1] from f[i], dir=-1 ==> compute f[i-1] from f[i] */
void POWIC_proceed (mpz_t *f, mpz_t *ff, int w, int dir){
  int i;
  mpz_t *fff = ff? ff: f;
  if ( dir==1 ){
    BLOOP (i, POWIC_quota, w){
      mpz_add (fff[i], fff[i], f[i-w]);
    }
  } else if ( dir==-1 ){
    FLOOP (i, w, POWIC_quota) mpz_sub (fff[i], fff[i], f[i-w]);
#ifdef POWIC_COUNT
    FLOOP (i, 0, POWIC_quota){
      if ( mpz_cmp_ui (fff[i], 0) == 0 ) POWIC_cnt++;
      POWIC_cnt_all++;
    }
#endif
  }
}


/* compute the sum of (f times (h[i]-h[j])) */
void POWIC_fh (mpz_t *s, mpz_t *f, mpz_t *b, int w){
  int i;
  mpz_t x, h;
  mpz_init (x); mpz_init_set_ui (h, 0);
  FLOOP (i, 0, POWIC_quota){
    mpz_add (h, h, b[i]);
    if ( i-w >= 0 ) mpz_sub (h, h, b[i-w]);
//gmp_printf ("%d   f[i]=%Zd   b[i]=%Zd   h= %Zd\n", i, f[POWIC_quota-1-i], b[i], h);
    mpz_mul (x, h, f[POWIC_quota-1-i]);
    mpz_add (*s, *s, x);
  }
  mpz_clear (x); mpz_clear (h);
}

/* compute the Banzhaf Index and return the values */
void POWIC_Banzhaf (mpz_t *v, int *w){
  int i, n=POWIC_players;
  mpz_t *f;

  mpz_init_set_ui (*v, 0);
  f = POWIC_mk_line (); mpz_set_ui (f[0], 1);   // initialization of f

  FLOOP (i, 0, n-1) POWIC_proceed (f, NULL, w[i], 1);  // forward DP

  FLOOP (i, POWIC_quota-w[n-1], POWIC_quota) mpz_add (*v, *v, f[i]); // compute the summation

  POWIC_clear_line (f);
}


/* compute the Shapley-Shubik Index and return the values */
void POWIC_Shapley_Shubik (mpz_t *v, int *w){
  int i, j, n=POWIC_players;
  mpz_t **f, x;

  mpz_init_set_ui (*v, 0); mpz_init_set_ui (x, 0);
  f = POWIC_mk_lines ();
  mpz_set_ui (f[0][0], 1);   // initialization (substitute 1 to emptyset)

  FLOOP (i, 0, n-1){
    BLOOP (j, i+1, 0) POWIC_proceed (f[j], f[j+1], w[i], 1);  // forward DP
  }

  FLOOP (i, 0, n-1){
    FLOOP (j, POWIC_quota-w[n-1], POWIC_quota){
       mpz_mul (x, f[i][j], POWIC_comb[i]);
       mpz_add (*v, *v, x); // compute the summation
    }
  }

  POWIC_clear_lines (f);
  mpz_clear (x);
}

/* compute the Deegan-Packel Index and return the values */
void POWIC_Deegan_Packel (mpz_t *v, int *w){
  int i, k, y, n=POWIC_players;
  mpz_t **f, x;
  
  mpz_init_set_ui (*v, 0); mpz_init_set_ui (x, 0);
  f = POWIC_mk_lines ();
  mpz_set_ui (f[0][0], 1);   // initialization (substitute 1 to emptyset)

    // 1. forward DP
  FLOOP (i, 0, n-1){
    BLOOP (k, i+1, 0) POWIC_proceed (f[k], f[k+1], w[i], 1);  // forward DP
  }

  FLOOP (k, 0, n-1){
    FLOOP (y, POWIC_quota-w[n-1], POWIC_quota){
       mpz_mul (x, f[k][y], POWIC_fact);
       mpz_div_ui (x, x, k+1);
       mpz_add (*v, *v, x); // compute the summation
    }
  }

  POWIC_clear_lines (f);
  mpz_clear (x);
}

/* compute the Banzhaf Index and return the values */
void POWIC_count_minimal_coalitions (int *w){
  int i, j;
  mpz_t *f, s;

  mpf_init2 (POWIC_minc, POWIC_prec);
  if ( POWIC_quota <= w[0] ){ mpf_set_ui (POWIC_minc, 1); return; };  // a player has majority
  f = POWIC_mk_line ();
  mpz_set_ui (f[0], 1); mpz_init_set_ui (s, 0);   // initialization of f
  FLOOP (i, 0, POWIC_players){
    FLOOP (j, POWIC_quota-w[i], POWIC_quota) mpz_add (s, s, f[j]);
    POWIC_proceed (f, NULL, w[i], 1);  // forward DP
  }
  mpf_set_z (POWIC_minc, s);
  mpz_clear (s);
  POWIC_clear_line (f);
}


/* compute the sum of (f times (h[i]-h[j])) */
void POWICD_fh (double *s, double *f, double *b, int w){
  int j;
  double h=0;
  FLOOP (j, 0, POWIC_quota){
    h += b[j];
    if ( j-w >= 0 ) h -= b[j-w];
    *s += h * f[POWIC_quota-1-j];
  }
}

/* compute the Banzhaf Index and return the values */
double *POWICD_Banzhaf (int *w){
  int i, y, n=POWIC_players;
  double *f, *b, *v;

  calloc2 (f, POWIC_quota+2, EXIT); f[0] = 1;   // initialization of f
  calloc2 (b, POWIC_quota+2, EXIT);
  calloc2 (v, POWIC_quota+2, EXIT);

  FLOOP (i, 0, n-1){
    BLOOP (y, POWIC_quota, w[i]) f[y] += f[y-w[i]];
  }

  BLOOP (i, n, 0){
    if ( i == n-1 ) b[0] = 1;  // initialization of b
    else {
      BLOOP (y, POWIC_quota, w[i+1]) f[y] += f[y-w[i+1]];
      FLOOP (y, w[i], POWIC_quota) f[y] -= f[y-w[i]];
    }
    v[w[i]] = 0;   // clear the variable
    if ( v[w[i]] == 0 ) POWICD_fh (&v[w[i]], f, b, w[i]);   // compute the summation
  }

  mfree (f, b);
  return (v);
}

/* compute the Shapley-Shubik(flag=0)/Deegan-Packel(flag=1) Index and return the values */
double *POWICD_SsDp (int *w, int flag){
  int i, k, y, n=POWIC_players;
  double **f, **b, *v;

  calloc2d (f, n+2, POWIC_quota+2, EXIT); f[0][0] = 1;
  calloc2d (b, n+2, POWIC_quota+2, EXIT);
  calloc2 (v, POWIC_quota+2, EXIT);

  FLOOP (i, 0, n-1){
    BLOOP (k, i+1, 0){
      BLOOP (y, POWIC_quota, w[i]) f[k+1][y] += f[k][y-w[i]];
    }
  }

  BLOOP (i, n, 0){
    if ( i == n-1){ 
      if ( !flag ) FLOOP (k, 0, i+1) b[k][0] = POWICD_comb[k];    // initialize b
    } else {
      FLOOP (k, 0, i+1){
        if ( flag) BLOOP (y, POWIC_quota+1, w[i]+1) b[k][y] = b[k][y-w[i]+w[i+1]] + b[k+1][y-w[i]];
        else BLOOP (y, POWIC_quota, w[i]) b[k+1][y] += b[k][y-w[i+1]];
      }
      FLOOP (k, 0, i+1){
        FLOOP (y, w[i], POWIC_quota) f[k][y] -= f[k+1][y-w[i]];
      }
    }
       // second case of the recursion of b
    if ( flag ) FLOOP (y, i==n-1? 1: w[i+1]+1, MIN (POWIC_quota, w[i]+1) ){
      FLOOP (k, 0, i+1) b[k][y] = ((double)1) /((double)(k+1));
    }

    if ( v[w[i]] == 0 ) FLOOP (k, 0, i+1){
      if ( flag ) FLOOP (y, 0, POWIC_quota) v[w[i]] += f[k][y] *b[k][POWIC_quota-y];
      else POWICD_fh (&v[w[i]], f[k], b[k], w[i]);   // compute the summation
    }
  }

  free2d (f); free2d (b);
  return (v);
}


/* compute the Shapley-Shubik Index and return the values */
double *POWICD_Shapley_Shubik (int *w){
  int i, k, y, n=POWIC_players;
  double **f, **b, *v;

  calloc2d (f, n+2, POWIC_quota+2, EXIT); f[0][0] = 1;
  calloc2d (b, n+2, POWIC_quota+2, EXIT);
  calloc2 (v, POWIC_quota+2, EXIT);

  FLOOP (i, 0, n-1){
    BLOOP (k, i+1, 0){
      BLOOP (y, POWIC_quota, w[i]) f[k+1][y] += f[k][y-w[i]];
    }
  }

  BLOOP (i, n, 0){
    if ( i == n-1){ 
      FLOOP (k, 0, i+1) b[k][0] = POWICD_comb[k];    // initialize b
    } else {
      FLOOP (k, 0, i+1){
        BLOOP (y, POWIC_quota, w[i]) b[k+1][y] += b[k][y-w[i+1]];
      }
      FLOOP (k, 0, i+1){
        FLOOP (y, w[i], POWIC_quota) f[k][y] -= f[k+1][y-w[i]];
      }
    }
    if ( v[w[i]] == 0 ) FLOOP (k, 0, i+1){
      POWICD_fh (&v[w[i]], f[k], b[k], w[i]);   // compute the summation
    }
  }

  free2d (f); free2d (b);
  return (v);
}

/* compute the Deegan-Packel Index and return the values */
double *POWICD_Deegan_Packel (int *w){
  int i, k, y, n=POWIC_players;
  double **f, **b, *v;
  
  calloc2d (f, n+2, POWIC_quota+2, EXIT); f[0][0] = 1;
  calloc2d (b, n+2, POWIC_quota+2, EXIT);
  calloc2 (v, POWIC_quota+2, EXIT);

  FLOOP (i, 0, n-1){
    BLOOP (k, i+1, 0){
      BLOOP (y, POWIC_quota, w[i]) f[k+1][y] += f[k][y-w[i]];
    }
  }

  BLOOP (i, n, 0){
    if ( i < n-1 ){
      FLOOP (k, 0, i+1){  // 1st case of recursion of b
        BLOOP (y, POWIC_quota+1, w[i]+1) b[k][y] = b[k][y-w[i]+w[i+1]] + b[k+1][y-w[i]];
      }
      FLOOP (k, 0, i+1){
        FLOOP (y, w[i], POWIC_quota) f[k][y] -= f[k+1][y-w[i]];
      }
    }
       // second case of the recursion of b
    FLOOP (y, i==n-1? 1: w[i+1]+1, MIN (POWIC_quota, w[i]+1) ){
      FLOOP (k, 0, i+1) b[k][y] = ((double)1) /((double)(k+1));
    }

    if ( v[w[i]] == 0 ) FLOOP (k, 0, i+1){  // compute the sum
      FLOOP (y, 0, POWIC_quota){
//gmp_printf ("i,k,y = %d,%d,%d, wi=%d,  %Zd  %Zd\n", i, k, y, w[i], f[k][y], b[k][POWIC_quota-y]);
        v[w[i]] += f[k][y] *b[k][POWIC_quota-y];
      }
    }
  }

  free2d (f); free2d (b);
  return (v);
}

/*************************************************************************/
/* main of POWIC */
/*************************************************************************/
int POWIC_main (int argc, char *argv[]){
  int *w, *ww, i;
  mpz_t v;
  mpf_t s, pow, fact;
    
    // parameter read and problem load
  POWIC_read_param (argc, argv);
  ARY_LOAD (w, int, POWIC_players, POWIC_infname, 1, EXIT);
    // compute the precision
  for (i=POWIC_players ; i>0 ; i/=2 ) POWIC_prec += POWIC_players;

    // compute quota
  if ( POWIC_quota == 0 ){
    ARY_SUM (POWIC_quota, w, 0, POWIC_players);
    POWIC_quota = (POWIC_quota+1) /2;
  }
  FLOOP (i, 0, POWIC_players) ENMIN (w[i], POWIC_quota);  // flooring too big weight

    // duplicate and sort the weight array
  ARY_DUP (ww, w, POWIC_players, EXIT);
  qsort_int (ww, POWIC_players, -1);

//  POWIC_test ();

     // power/combination numbers
  malloc2 (POWIC_comb, POWIC_players+1, EXIT);
  mpz_init_set_ui (POWIC_comb[0], 1);
  mpf_init2 (pow, POWIC_players+2); mpf_set_ui (pow, 1); 
  mpf_init2 (fact, POWIC_prec); mpz_init_set_ui (POWIC_fact, 1);
  FLOOP (i, 1, POWIC_players+1){
    mpf_mul_ui (pow, pow, 2);    // compute 2^n
    mpz_mul_ui (POWIC_fact, POWIC_fact, i);  // compute n!
    if ( i<POWIC_players ) mpz_mul_ui (POWIC_comb[0], POWIC_comb[0], i);  // comb[0] := (n-1)!
  }
  mpf_set_z (fact, POWIC_fact);
  FLOOP (i, 1, POWIC_players){
    mpz_init (POWIC_comb[i]);
    mpz_mul_ui (POWIC_comb[i], POWIC_comb[i-1], i);
    mpz_div_ui (POWIC_comb[i], POWIC_comb[i], POWIC_players-i);  // comb[k] := k!(n-1-k)!
  }
  
// main routines
  if ( POWIC_problem & POWIC_BZ ) POWIC_Banzhaf (&v, ww);
  else if ( POWIC_problem & POWIC_SS ) POWIC_Shapley_Shubik (&v, ww);
  else {
    POWIC_count_minimal_coalitions (ww);
    POWIC_Deegan_Packel (&v, ww);
  }
#ifdef POWIC_COUNT
printf ("DP end\n");
#endif
   // compute and print fraction-ed values
  mpf_init2 (s, POWIC_prec);
/*
  FLOOP (i, 0, POWIC_players){
    mpf_set_z (s, v[w[i]]);
    mpf_div (s, s, (POWIC_problem&POWIC_BZ)? pow: fact);
    if ( POWIC_problem & POWIC_DP ) mpf_div (s, s, POWIC_minc);
#ifdef POWIC_COUNT
    gmp_printf("%Zd %10.10Ff\n", v[w[i]], s);
#else
    gmp_printf("%1.20Ff\n", s);
#endif
    if ( mpf_cmp_ui (s, 1) > 0 ){ printf ("okasii\n"); exit(0); }
    mpz_clear (POWIC_comb[i]);
  }
#ifdef POWIC_COUNT
  printf ("ratio of 0-cell %lld / %lld  (%lld%%)\n", POWIC_cnt, POWIC_cnt_all, POWIC_cnt*100/POWIC_cnt_all);
#endif

    // clear the variables
  POWIC_clear_line (v);
  mfree (POWIC_comb, w, ww);
  mpz_clear (POWIC_fact); mpf_clear (pow); mpf_clear (fact); mpf_clear (s);
  if ( POWIC_problem & POWIC_DP ) mpf_clear (POWIC_minc);
*/

  mpf_set_z (s, v);
  mpf_div (s, s, (POWIC_problem&POWIC_BZ)? pow: fact);
  gmp_printf("%1.20Ff\n", s);
  mpz_clear (v);  mpf_clear (s);
  return (ERROR_MES?1:0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (POWIC_main (argc, argv));
}
#endif
/*******************************************************************************/

#endif


